/*
 *
 * mknod /dev/mv1000 		c 26 0
 *
 * Copyright (C)  1995-1997 2000 Jochen Karrer, LinuxData GbR 
 *
 * Released under the terms of the GPL.
 *
 */

/* Kernel includes */

#include "linux/config.h"
#include "linux/module.h"
#include <linux/version.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/major.h>
#include <linux/poll.h>
#include <asm/segment.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/signal.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/ipc.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include "mv1drv.h"
#include "mv1regs.h"
#include <asm/io.h>
#include <asm/pgtable.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/proc_fs.h>

MODULE_AUTHOR("Jochen Karrer, LinuxData GbR, Germany");
MODULE_DESCRIPTION("Driver for MuTech MV1000 Framegrabber");

#if (LINUX_VERSION_CODE<KERNEL_VERSION(2,4,0)) || (LINUX_VERSION_CODE>=KERNEL_VERSION(2,5,0))
#error This driver is for Linux Kernel 2.4 
#endif

static int mv1_numboards  = 0;
#define MAXBOARDS  3 	// Support for more than one board per driver is totally broken

MV1000 *mv1_board[MAXBOARDS]={0,};


static int lpt_port_base = 0x0;
static unsigned char integration_mode=0;
static int init_module_done=0;

static int emulate_irq = 0;
MODULE_PARM(emulate_irq,"i");
MODULE_PARM(lpt_port_base,"i");
				// empty space mode

static void mv1_send_event(MV1000 *mv1,MV1_Event ev);
/*
 * The driver.
 */
#include "mv1drv_io.c"

static inline void 
ccd_stop_integration(MV1000 *bs)
{
	if(integration_mode<2) {
		mv1_set_gp_bits(0^integration_mode,bs); // stop integration
		if(lpt_port_base)
			outb(255*(1^integration_mode),lpt_port_base);
	 }
}

static inline void 
ccd_start_integration(MV1000 *bs)
{
	if(integration_mode<2) {
		mv1_set_gp_bits(1^integration_mode,bs); // start integration 
		if(lpt_port_base)
			outb(255*(0^integration_mode),lpt_port_base);
	}
}


/*
 * -------------------------------------------------------------------------------
 *
 * mv1_frame_end_interrupt --
 *
 * 	Starts and stops the integration on the CCD-Chip of my Pulnix Camera	
 *
 * -------------------------------------------------------------------------------
 */
static inline void
mv1_frame_end_interrupt(MV1000 *bs ){
       	kill_fasync(&bs->fasync, SIGIO,POLL_IN);
	if ( bs->aq_frame ){
		if(bs->rest_time!=1){ 
			ccd_start_integration(bs);
		}

		bs->do_transfer=1;
		if(bs->frame_end_handler) {
			bs->frame_end_handler(&bs->curr_gw,bs->aqdata);
		}

		bs->frame_count++;
		bs->aq_frame = 0;
		bs->snap_flag = 0;
		wake_up_interruptible(&bs->frame_end_wait);
	}
	bs->rest_time--;
	if(bs->rest_time == 1) {
		// enable grab on next frame ( hack for 10 Bit Version )
		mv1_set_grab_start_addr(bs->curr_gw.startaddr,bs);
	}
	if(bs->rest_time <1) {
		ccd_stop_integration(bs);
		if(bs->rest_time == 0) {
			bs->aq_frame = 1;	// queue aquisition for next frame
		}
		if((!bs->snap)|| (bs->snap_flag && (bs->aq_frame == 0) ) ) {
			bs->rest_time = bs->i_time;
		}
		if(bs->rest_time != 1)
			// disable grab  on next frame by setting to non existing mem.
			mv1_set_grab_start_addr(3*1024*1024,bs); // hack for 10 Bit Version
		else 
		 	mv1_set_grab_start_addr(bs->curr_gw.startaddr,bs);

	}
}

int 
mv1_delete_frame_end_handler() 
{
	MV1000 *mv1=mv1_board[0];		
	if(!mv1)
		return -ENODEV;
	mv1->frame_end_handler = NULL;
	mv1->aqdata=NULL;
	return 0;
}

int 
mv1_create_frame_end_handler(void (*feh)(MV1_GrabWindow *gw,void *aqdata),void *aqdata) 
{
	MV1000 *mv1=mv1_board[0];		
	if(!mv1)
		return -ENODEV;
	mv1->aqdata=aqdata;
	mv1->frame_end_handler = feh;
	return 0;
}


static void 
mv1_interrupt(int irq,void *dev,struct pt_regs *regs) {
	MV1000 *bs = (MV1000 *) dev;
	u_long pnd;
	static u_long maxints=0;

	pnd = mv1_clear_all_irqs(bs);
	
	if(is_transfer_end_irq(pnd)) {
		bs->transfer_end_count++;
		if(!mv1_inquire_master_loop(bs))
			mv1_set_master_enable(0,bs);  // or crash
		if(bs->do_transfer) {
//			do_rheed_aquisition(&bs->curr_gw,bs->aqdata);
			wake_up_interruptible(&bs->transfer_end_wait);
			bs->do_transfer=0;
		}
	}
	if(is_frame_end_irq(pnd)) {
		mv1_frame_end_interrupt(bs);
		bs->frame_int_count++;
	}
	if(is_bad_list_irq(pnd)) {
		mv1_set_master_loop(0,bs);
		mv1_set_master_enable(0,bs); 
		mv1_set_transfer_end_irq(0,bs);
		mv1_send_event(bs,MV1_SET_MASTER_ENABLE);
		mv1_send_event(bs,MV1_SET_TRANSFER_END_IRQ);
		bs->bad_list_count++;
	}
	if(jiffies == bs->last_interrupt){
		maxints++;
		if(maxints > 8 ) 
			goto stop_interrupt;
	}
	else {
		maxints = 0;
	}
	bs->last_interrupt = jiffies;
	sti();
	return;

stop_interrupt:
	mv1_force_clear_irqs(bs);	
	mv1_set_transfer_end_irq(0,bs);
	mv1_set_frame_end_irq(0,bs); 
	mv1_set_master_enable(0,bs); 
	free_irq(irq,bs); // should mask the interrupt in PIC 
			  // if interrupts will still arrive.
			  // see handler for bad_interrupt in irq.c
	bs->irq = 0;
}

/*
 * -----------------------------------------------------------------
 *
 *  Emulation of frame_end interrupt if no framegrabber is present
 *
 * -----------------------------------------------------------------
 */
static void mv1_emulate_interrupt(unsigned long);

static struct timer_list  emutimer = { data:0, function: mv1_emulate_interrupt };
static int emutimer_busy=0;

static void 
mv1_emulate_interrupt(unsigned long data) {
	mv1_frame_end_interrupt(mv1_board[0]);
	init_timer(&emutimer);
        emutimer.expires=jiffies+4; 
	add_timer(&emutimer);
	emutimer_busy=1;
}

static void
mv1_watchdog(unsigned long data) {
	MV1000 *mv1 = (MV1000*)mv1;
        mv1->watchdog_timer.expires=jiffies+HZ; 
	add_timer(&mv1->watchdog_timer);
	mv1->watchdog_timer_busy =1;
	if(jiffies - mv1->last_interrupt > HZ/2) {
                if( mv1_inquire_frame_end_irq(mv1) &&
                        mv1_inquire_grab(mv1) && mv1_inquire_grabflag(mv1) ) {
                        mv1_set_grab(0,mv1);
                        mv1_set_grab(1,mv1);
                }
 
        }      
}
/*
 * ----------------------------------------------------------------------------
 *
 * mv1_read -- 
 *
 *	reading from /dev/mv1000_ctrl gives data from the event-queue
 *	reading from /dev/mv1000 gives the contents of the current frame
 *
 * ----------------------------------------------------------------------------
 */

static inline ssize_t 
mv1_event_read(struct file * file,char * buf,size_t count,loff_t* nfpos)
{
	int i;
	mv1_file_data *pd = file->private_data;
	MV1000 *mv1 = pd->mv1;
	for(i=0;i<count-(sizeof(MV1_Event)-1);i+=sizeof(MV1_Event)){
		MV1_Event event;
		while(mv1->event_queue_wp == pd->event_rp) {
			if(i) 
				return i;
			if(file->f_flags & O_NONBLOCK) {
                                 return -EWOULDBLOCK;
                        }	
			interruptible_sleep_on(&mv1->event_wait);
			if (signal_pending(current))  
				return -ERESTARTSYS;
		}
		event = mv1->event_queue[pd->event_rp];
		put_user(event,(MV1_Event *)buf);
		mv1->event_tab[event] = 0;
		buf+=sizeof(MV1_Event);
		pd->event_rp = (pd->event_rp + 1) % MV1_EVENTQUEUE_SIZE;
		file->f_pos+=sizeof(MV1_Event); 
	}
	return i;

}

static ssize_t
mv1_image_read (struct file * file,char * buf,size_t count,loff_t* nfpos)
{
	mv1_file_data *pd = file->private_data;
	MV1000 *bs = pd->mv1;
	unsigned long fpos = (unsigned long) file->f_pos;
	unsigned long page;	
	unsigned long ofs;
	unsigned long cnt;
	unsigned int imgsize = bs->curr_gw.dx*bs->curr_gw.dy;

	pd->last_read_frame = bs->frame_count;
	if(fpos>=imgsize)
		return 0;
	if(fpos+count > imgsize) 
		count = imgsize-fpos;

	cnt=count;
	while(cnt) {
		int len;
		int maxlen;
		len = cnt;
		page = fpos >> PAGE_SHIFT;
		ofs  = fpos & (PAGE_SIZE-1);
		maxlen = PAGE_SIZE - ofs;
		if(len>maxlen)
			len=maxlen;
		if(page>MASTERMEM_PAGES)
			return -EINVAL;
		if(!bs->mastermem[page]) {
			return -EINVAL;	
		}
		copy_to_user(buf,(void*)bs->mastermem[page]+ofs,len);
		buf+=len;
		fpos+=len;
		cnt-=len;
	}
	file->f_pos+=count;
	return count;
}

static ssize_t 
mv1_read(struct file * file,char * buf,size_t count,loff_t* nfpos)
{
	struct inode *inode = file->f_dentry->d_inode;

	switch(MINOR(inode->i_rdev)){
	    
	   case MV1_CONTROL_MINOR:
		return mv1_event_read(file,buf,count,nfpos);
	   case MV1_RAW_MINOR:
		return mv1_image_read(file,buf,count,nfpos);
	   default:
		return -ENODEV;
	}
}

/*
 * ------------------------------------------------------------------------------
 *
 * mv1_write --
 * 
 *	write one dataset for every screen shot to the device driver.
 *
 * ------------------------------------------------------------------------------
 */
 
static ssize_t 
mv1_write(struct file * file,const char * buf,size_t count,loff_t *fpos)
{
	struct inode *inode = file->f_dentry->d_inode;

	switch(MINOR(inode->i_rdev)) {
	    case MV1_RAW_MINOR:  // commands	
			return -ENODEV;
 	    default:
		return -ENODEV;
	}
}

/*
 * ----------------------------------------------------------------------------------
 *
 * mv1_poll --
 *
 *	mv1_control_minor:
 *		add mv1_event_wait to select table
 *
 * ---------------------------------------------------------------------------------- 
 */

static unsigned int 
mv1_raw_poll(struct file *filp,poll_table *wait){
	mv1_file_data *pd = filp->private_data;
	MV1000 *mv1 = pd->mv1;
	unsigned int mask=0;

	poll_wait(filp,&mv1->transfer_end_wait,wait);
	if(mv1->frame_count > pd->last_read_frame) {
       		mask |= POLLRDNORM | POLLIN;
	}
	return mask;
}

static unsigned int 
mv1_control_poll(struct file *filp,poll_table *wait){
	mv1_file_data *pd = filp->private_data;
	MV1000 *mv1 = pd->mv1;
	unsigned int mask=0;
	poll_wait(filp,&mv1->event_wait,wait);
	if(mv1->event_queue_wp != pd->event_rp) {
       		mask |= POLLRDNORM | POLLIN;
	}
	return mask;
}

static unsigned int 
mv1_poll(struct file *filp,poll_table *wait){
	struct inode *inode = filp->f_dentry->d_inode;
	switch(MINOR(inode->i_rdev)) {
		case MV1_CONTROL_MINOR:
			return mv1_control_poll(filp,wait);

		case MV1_RAW_MINOR:
			return mv1_raw_poll(filp,wait);
		default:
			return -ENODEV;
	}	
}

static void
mv1_send_event(MV1000 *mv1,MV1_Event ev){
        if((jiffies - mv1->event_tab[ev]) > HZ){
                mv1->event_queue[mv1->event_queue_wp] = ev;
                mv1->event_tab[ev] = jiffies;
                mv1->event_queue_wp = ( mv1->event_queue_wp + 1 ) % MV1_EVENTQUEUE_SIZE;
                wake_up_interruptible(&mv1->event_wait);
		// send_sig to all tasks using the mv1000
       	} 
}


/*
 * -------------------------------------------------------------------
 *
 * some convenience functions  for ioctl, needed for copying 
 * arguments between  kernel and userspace.
 *
 * -------------------------------------------------------------------
 */
static int 
ioctl_get_arg(u_long *arg1,u_long *argv){
        get_user(*arg1,argv);
        return 0;
}

static int 
ioctl_get_2_args(u_long *arg1,u_long *arg2,u_long * argv){
	get_user(*arg1,argv);
	get_user(*arg1,argv+1);
	return 0;
}

static int 
ioctl_put_arg(u_long arg,u_long *argv){
	put_user(arg,argv);
	return 0;
}

static int 
mv1_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
  	mv1_file_data *pd = file->private_data;
	MV1000 *bs = pd->mv1;
	MV1_GrabWindow *gw = &bs->curr_gw;
	int i;
	
	switch(MINOR(inode->i_rdev)) {
	    case MV1_CONTROL_MINOR:
		switch (cmd) {

		    case MV1_EVENT_SEND:
			mv1_send_event(bs,arg);
			return 0;

		    #include "mv1_ioctl.c"
		    default:
			return -EINVAL;
		}	
	  case MV1_RAW_MINOR:
		switch(cmd){
			case MV1_SET_GRAB_WINDOW:
        			copy_from_user(gw,(void*)arg,sizeof(MV1_GrabWindow));
				mv1_recalc_curr_gw(bs);
				return 0;

			case MV1_INQUIRE_GRAB_WINDOW:
                                copy_to_user((void*)arg,gw,sizeof(MV1_GrabWindow));
                                return 0;

			case MV1_SET_FRAME:
                                copy_from_user(&(pd->frame),(void*)arg,sizeof(MV1_Frame));
                                return 0;

			case MV1_INQUIRE_FRAME:
                                copy_to_user((void*)arg,&(pd->frame),sizeof(MV1_Frame));
                                return 0;

	   	}
	   default:
		return -ENODEV;
	}
	
	return -EINVAL;
}

/*
 * support seeks on the device
 */


static loff_t 
mv1_llseek(struct file * file, loff_t offset, int orig)
{
	struct inode *inode = file->f_dentry->d_inode;
  	mv1_file_data *pd = file->private_data;
	MV1000 *mv1 = pd->mv1;
    	switch(MINOR(inode->i_rdev)) { 
		case MV1_RAW_MINOR:
			switch (orig) {
				case 0:	// seek_set 
					file->f_pos = offset;
					return file->f_pos;
				case 1: // seek_cur 
					file->f_pos += offset;
					return file->f_pos;
				case 2:	// seek_end 
					pd->last_read_frame = mv1->frame_count; 
					return file->f_pos;
				default:
					return -EINVAL;
			}

		default: 
			return -ENODEV;
	}
}
/*
 * ------------------------------------------------------------------
 * for security: destroy all sdc's when freeing any master-memory 
 * ------------------------------------------------------------------
 */ 
static void
mv1_destroy_sdcs(MV1000 *mv1){
        int i;

        mv1_set_grab_mask(0,mv1);
        mv1_set_host_mask(1,mv1);
        mv1_set_mem_mask(~0,mv1);
	mv1->tlist_valid=0;
        for(i=0;i<65536;i+=4){
		if(!mv1->simulation)
                	writel(11,mv1->vvmem+i);
                readl(mv1->vvmem+i);
        }
        mv1_set_host_mask(0,mv1);
        mv1_set_grab_mask(1,mv1);
}


/*
 * ------------------------------------------------------------------
 *
 * mv1_mmap --
 *
 *	raw_minor:
 *
 *	nearly the same as mmap from /dev/mem ,
 *	but its only possible to map the master transfered image 
 *
 *	control_minor:
 *	
 *	map the registers and the mv1 onboard memory to userspace
 *	(only needed for driver development)
 *
 * -------------------------------------------------------------------
 */

static int 
mv1_mmap(struct file * file, struct vm_area_struct * vma)
{
	struct inode *inode = file->f_dentry->d_inode;
	int size;
	mv1_file_data *pd = file->private_data;
	MV1000 *bs = pd->mv1;

        size = vma->vm_end - vma->vm_start;

	switch(MINOR(inode->i_rdev)) {
#ifdef DEBUG
            	case MV1_CONTROL_MINOR:	{
			unsigned long offset = vma->vm_pgoff<<PAGE_SHIFT;
			if( (offset + size > 8*1024*1024 ) 
				|| (! bs->phys_mem ) ) {
				return -EPERM;
			}
			offset += bs->phys_mem;
			pgprot_val(vma->vm_page_prot) |= _PAGE_PCD | _PAGE_PWT;
			vma->vm_flags |= VM_IO;
			if (remap_page_range(vma->vm_start, offset,
				 size, vma->vm_page_prot))
				return -EAGAIN;
			return 0;
		}
#endif

		case MV1_RAW_MINOR: {
			unsigned long start= vma->vm_start;
        		unsigned long page;
			int i;
                        if ((size>>PAGE_SHIFT)>MASTERMEM_PAGES)
                		return -EINVAL;
        		for(i=0;i<(size>>PAGE_SHIFT);i++)  {
                		page = bs->mastermem[i];
				if(!page)
					break;
                		if (remap_page_range(start, __pa(page), PAGE_SIZE, PAGE_SHARED))
                        		return -EAGAIN;
                		start+=PAGE_SIZE;
        		}
        		return 0;            
		}
		default:
	    		return -EINVAL;
	
	}
}


static int
mv1_open( struct inode* inode, struct file* file)
{
	mv1_file_data *pd;
	MV1000 *mv1 = mv1_board[0];
	if(!init_module_done) {
		printk("mv1000: called open before init_module_done=1\n");
	}
	if(!(file->private_data = 
	    kmalloc(sizeof(mv1_file_data),GFP_KERNEL)))
		return -ENOMEM;

	/*	initialize  private_data field */

	pd = (mv1_file_data *) file->private_data;	
	pd->mv1 = mv1;

	switch(MINOR(inode->i_rdev)) {
		case MV1_CONTROL_MINOR:
			pd->event_rp = mv1->event_queue_wp ;
		  	break;
		
		
		case MV1_RAW_MINOR:
			pd->frame.sx =  0;
			pd->frame.sy =  0;
			pd->frame.dx =  mv1->curr_gw.dx; // 768;
			pd->frame.dy =  mv1->curr_gw.dy; // 574;
			pd->last_read_frame = 0;
			break;
		default:
			return -ENODEV;
	}
	MOD_INC_USE_COUNT;
	return 0;   
}


static int
mv1_fasync(int fd, struct file *filp, int on)
{
	mv1_file_data *pd = filp->private_data;
	MV1000 *mv1 = pd->mv1;
        int retval;
        retval = fasync_helper(fd, filp, on, &mv1->fasync);
        if (retval < 0)
                return retval;
        return 0;
}


static int 
mv1_release(struct inode* inode, struct file* filp)
{
	mv1_file_data *pd = filp->private_data;
	mv1_fasync(-1,filp,0);
	kfree(pd);
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations mv1_fops = {
	llseek:		mv1_llseek,
	read:		mv1_read,
	write:		mv1_write,
	poll:		mv1_poll, //mv1_select,	/* mv1000_select */
	ioctl:		mv1_ioctl,
	mmap:		mv1_mmap,	/* mv1000_mmap  */
	open:		mv1_open,
	release:        mv1_release,
	fasync:		mv1_fasync,
};

/*
 * -------------------------------------------------------------------
 *
 *  handle_triton --
 *
 *	there is a hardware problem with the mv1000 and the 82437 
 *	chipset. this function disables streaming. in combination
 *	with a high master-latency timer on the mv1000 i get 
 *	about 30 mb/s, but still crashes.
 *
 * ------------------------------------------------------------------
 */

#define  TRITON_PCON			(0x50)
#define  TRITON_BUS_CONCURRENCY		(1<<0)
#define  TRITON_STREAMING		(1<<1)
#define  TRITON_PEER_CONCURRENCY	(1<<3)
#define  TRITON_WRITE_BURST		(1<<2)

static void
handle_triton(){

	int pci_index=0;

	unsigned char pci_bus, pci_device_fn;
	u_char b;
	if (pcibios_find_device (
		PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437,
		pci_index, &pci_bus, &pci_device_fn) == 0) {
                        printk("found 82437\n");
                        pcibios_read_config_byte(pci_bus,pci_device_fn,TRITON_PCON,&b);
                        printk("triton pcicon: 0x%x\n",b);
                        if(!(b&TRITON_STREAMING))
                                printk("82437fx: disabling streaming\n");
                        if(!(b&TRITON_BUS_CONCURRENCY)) 
                                printk("82437: disabling bus concurrency\n");
                        if(b&TRITON_PEER_CONCURRENCY) 
                                printk("82437: disabling peer concurrency\n");
                        b |=  TRITON_STREAMING | TRITON_BUS_CONCURRENCY;
                        b &= ~TRITON_PEER_CONCURRENCY;
                        pcibios_write_config_byte(pci_bus,pci_device_fn,TRITON_PCON,b);
                }
                if(pcibios_find_device(PCI_VENDOR_ID_INTEL,PCI_DEVICE_ID_INTEL_82439,
                    pci_index,&pci_bus,&pci_device_fn)==0){
                        printk("found 82439\n");
                        pcibios_read_config_byte(pci_bus,pci_device_fn,TRITON_PCON,&b);
                        printk("triton pcicon: 0x%x\n",b);
                        if(b&TRITON_PEER_CONCURRENCY) {
                                printk("82439: disabling peer concurrency\n");
                                b &= ~TRITON_PEER_CONCURRENCY;
                        }
                        pcibios_write_config_byte(pci_bus,pci_device_fn,TRITON_PCON,b);

                }
}

static int
mv1_add_to_transfer_list(MV1000 *mv1,u32 src,u32 dst,int len,MV1_TransferList *tl) {
        if(src+len >= 4 *1024*1024) {
		return 0;
        }
        while(len > 0) {
        	int tlen;
        	int rest;
        	int srest;
        	int maxtransfer;
		unsigned int page;
        	transf_dscr *td = &tl->td[tl->len];

		page = dst >> PAGE_SHIFT;
		if(page>MASTERMEM_PAGES) {
			return 0;
		}
                rest = dst & (PAGE_SIZE-1);
                srest = src & (4095);  
                maxtransfer = PAGE_SIZE - rest;
		if((4096-srest)<maxtransfer) {
			maxtransfer = 4096-srest;
		}
                if(len > maxtransfer)
                        tlen = maxtransfer;
                else
                        tlen = len;
                td->src = src;
                td->dst = virt_to_bus((void*)mv1->mastermem[page])+rest;
		td->len = (((tlen&~3)^0xfff)+1)<<14;
		td->cksum = ~(td->len + td->src + td->dst) +1;
                tl->len++;
                len-=tlen; src += tlen; dst += tlen;
                if(tl->len > 2047) {
			return 0;
                }
        }
        return tl->len;
}

/*
 * --------------------------------------------------------
 * Write to MV1000 onboard memory
 * needs a special function because PCI-Bus will hang when
 * a 32 Bit boundary is crossed 
 * --------------------------------------------------------
 */
static int 
mv1_memwrite(MV1000 *mv1,void *d, void *s,int len) {
	unsigned long n;
	if((len&3) || ((unsigned long)s&3) || ((u32)d&3)) {
		printk("mv1000: ignore unaligned write to onboard memory !\n");
		return -1;
	}
        for(n=0;n<len;n+=4) {
		u32 val;
		writel(*((u32 *)(s+n)),mv1->vvmem+(d+n));
		val = readl(mv1->vvmem+d+n); // or crash (flush the bridge)
		if(val != *((u32*)(s+n))) {
			printk("mv1_memwrite error !\n");
			return -1;
		}
        }
        for(n=0;n<len;n+=4) {
		u32 val;
		val = readl(mv1->vvmem+d+n); 
		if(val != *((u32*)(s+n))) {
			printk("mv1_memcmp error !\n");
			return -1;
		}
	}
	return 0;
}

static void 
mv1_clear_transfer_list(MV1_TransferList *tl) {
	memset(tl,0,sizeof(*tl));
}

static int
mv1_write_transfer_list(MV1000 *mv1, MV1_TransferList *tl,int nr) {
	int mode;
	int error=0;
//	unsigned int flags;
        transf_dscr td_end = {0,};
	mv1->tlist_valid=0;
	td_end.len = (((0&~3)^0xfff)+1)<<14;

	if((nr>1) || (tl->len > 2047) )
		return -1;

//	save_flags(flags);
	mode = mv1_inquire_master_enable(mv1);
        mv1_set_master_enable(0,mv1);
	udelay(5);
	mv1_destroy_sdcs(mv1);
        mv1_set_grab_mask(0,mv1);
	mv1_set_host_mask(1,mv1);
	udelay(20);
	error += mv1_memwrite(mv1,(void*)(nr*32768),tl->td,tl->len*sizeof(transf_dscr));
	error += mv1_memwrite(mv1,(void*)(nr*32768+tl->len*sizeof(transf_dscr)),
                       &td_end,sizeof(transf_dscr));
	udelay(5);
	mv1_set_host_mask(0,mv1);
	udelay(5);
	mv1_set_grab_mask(1,mv1);
//	restore_flags(flags);
	if(error==0) { 
		mv1->tlist_valid=1;
        	mv1_set_master_enable(mode,mv1);
	}
        return error;
}

static void 
mv1_create_transfer_list(MV1000 *mv1,MV1_GrabWindow *gw,MV1_TransferList *tl) 
{
	mv1_clear_transfer_list(tl);
	if(!gw->interlaced) {
		mv1_add_to_transfer_list(mv1,gw->startaddr,0,gw->dx*gw->dy,tl);
	} else {
		unsigned long f_start;
		unsigned long s_start;
		int y;
		if(gw->start_even) {
			f_start = gw->startaddr; 
                        s_start = f_start + gw->field_offset;          
		} else {
			s_start = gw->startaddr;                                        
			f_start = s_start + gw->field_offset;          
		}
		for(y=0;y<gw->dy;y++) {
			if(y&1) {
				mv1_add_to_transfer_list(mv1,s_start,y*gw->dx,gw->dx,tl);
				s_start+=gw->dx;
			} else {
				mv1_add_to_transfer_list(mv1,f_start,y*gw->dx,gw->dx,tl);
				f_start+=gw->dx;
			}
		}
	}
	mv1_write_transfer_list(mv1,tl,0);
}

static void
mv1_recalc_curr_gw(MV1000 *bs)
{
        MV1_GrabWindow *gw = &bs->curr_gw;
	if((H_END(bs) < H_START(bs)) || (bs->v_end < bs->v_start)) {
		printk("mv1000: gw size smaller than zero\n");
		return;
	}
        bs->curr_gw.sx = H_START(bs);
        bs->curr_gw.sy = bs->v_start;
	bs->curr_gw.dx = H_END(bs) - H_START(bs);
	bs->curr_gw.dy =  bs->v_end -  bs->v_start;
	bs->curr_gw.field_offset = bs->curr_gw.dx * bs->curr_gw.dy;
	if(gw->start_even) {
		gw->f_start = gw->startaddr + bs->vvmem;
		gw->s_start = gw->f_start + gw->field_offset;
	}
	else {
		gw->s_start = gw->startaddr + bs->vvmem;
		gw->f_start = gw->s_start + gw->field_offset;
	}
	if(bs->curr_gw.interlaced)
		bs->curr_gw.dy <<=1;

	if(bs->curr_gw.dy*bs->curr_gw.dx > 20000) {
		mv1_create_transfer_list(bs,&bs->curr_gw,bs->tlist[0]);
	} else {
		printk("mv1000: grab window to small. ignoring !\n");
	}
}                     

static void 
mv1_init_lm1882(MV1000 *bs) {
	int i;
	static short lm[]=
		       { 0x400, 	// status register
			 0x17, 		// hfp end time	
			 0x5b, 		// hsync pulse end time
			 0x9d, 		// hblank pulse end time
			 0x38e,		// total horizontal clocks
			 0x7, 		// vfp end time
			 0xd, 		// vsync pulse end time
			 0x29, 		// vblank pulse end time
			 0x20d, 	// total vertical lines
			 0x39,		// equalization pulse end time
			 0x19a,		// serration pulse start time
			 0x1, 		// pulse interval start time
			 0x13,		// pulse interval end time
			 0x29,		// vertical interrupt activate time
			 0x20e,		// vertical interrupt deactivate time
			 0x38f,		// horizontal drive start time
			 0x5c,		// horizontal drive end time
			 0x1,		// vertical drive start time
			 0x15  	};	// vertical drive end   time


	for(i=0;i<19;i++) {
		lm1882_out(lm[i],i,bs);
	}
}
/*
 * -----------------------------------------------------------------------
 *
 *  mv1_init_pll --
 *
 *	load the registers of the ics1522 with default values
 *
 * ----------------------------------------------------------------------- 
 */
static void 
mv1_init_pll(MV1000 *bs) {
	int i;
	bs->icsreg[0] = 908-1; 	// default for my pulnix tm-765 plt
        bs->icsreg[3] = 0 + (1<<10); // pll-ref-polarity negative 
        bs->icsreg[4] = 4 | (3<<3) | PDEN | INT_VCO | CLK_SEL | (1<<10);
        bs->icsreg[5] = FBK_SEL | (3 << 4) | (3<<6) | LD_LG | F_EN |(1<<10);
        bs->icsreg[6] = OMUX1 |OMUX3| EXTREF | 7 ;
	for(i=0;i<7;i++){
		mv1_ics_out(bs->icsreg[i],i,bs);
	}
	
}
/*
 * ----------------------------------------------------------------------
 *
 * mv1_init_dacs
 *
 *	load the da-convertes with default values
 *
 * ----------------------------------------------------------------------
 */
static void
mv1_init_dacs(MV1000 *bs){
	int i;
	bs->dac[0] = 0; 	// pll - phase trim
	bs->dac[1] = 102;	// ad-top
	bs->dac[2] = 31;	// ad-bottom
	bs->dac[3] = 61;	// clamp level
	bs->dac[4] = 0;	// dc-offset
	for(i = 0;i<5;i++){
		mv1_dac_out(bs->dac[i],i,bs);
	}
}
/*
 * ---------------------------------------------------------------------
 *
 *  mv1_init_xilinx
 *
 *	init the registers in the xilinx fpga
 *
 * ---------------------------------------------------------------------
 */

static void 
mv1_init_xilinx(MV1000 *bs) {
		bs->vctl = (8 << 12)+ (2<<8) + ( 8 << 3);
		bs->config_reg = mv1_config_inw(bs);
                bs->video_ctl =  CAPTURE_ENABLE_MASK | 
                        CLAMP_ON_BPORCH_MASK | INTERLACE_MASK |
                        EQ_REMOVE_MASK |EVEN_START_MASK 
			| MSTR_INT_PND_MASK | VID_INT_PND_MASK | 
			MSTR_BAD_LIST_MASK;

		bs->eq_remove = 24;
                bs->misc_ctl = bs->eq_remove << 26; 
                bs->h_start = 106;
                bs->v_start = 20;
                bs->h_v_start = MSTR_WAIT_FRAME_MASK | VIDEO_MASK_MASK | 
				( (106 >> 2) << 16) | (20 << 2);
                bs->h_end = 874;
                bs->v_end = 308;
		bs->h_v_end = (( 874 >> 2) << 16) | (308 << 2 );
                bs->mem_mask = ~0l;
		if(mv1_is_10_bit(bs)){
			bs->vctl |= SHIFT_DOWN_2_MASK;
		}
		mv1_extoutl(bs->h_v_start,H_V_START,bs);
		mv1_extoutl(bs->h_v_end,H_V_END,bs);
		mv1_extoutl(bs->misc_ctl,MISC_CONTROL,bs);
		mv1_vctl_outw(bs->vctl,bs);
		mv1_extoutl(bs->video_ctl,VIDEO_CONTROL,bs);
		bs->grab_start_addr = 65536; // room for master-sdc
		mv1_extoutl(bs->grab_start_addr,GRAB_START_ADDR,bs);
	
}
/*
 * ----------------------------------------------------------------------------
 *
 * mv1_init_filter --
 *
 *	enable filter ( switch of power down-control )
 *	set default values
 *
 * ----------------------------------------------------------------------------
 */
static void
mv1_init_filter(MV1000 *bs) {
	bs->filter_pdc = 0;
	bs->freq_cutoff = 127;
	bs->boost = 0;
	mv1_32f8102_out(bs->filter_pdc,SSI_PDC,bs); 	// enable filter
	mv1_32f8102_out(bs->freq_cutoff,SSI_DACF,bs);
	mv1_32f8102_out(bs->boost,SSI_DACB,bs); 
}
static void
mv1_init_luts(MV1000 *bs) {
	int i;
	bs->lut_ctl =0;
	for(i=7;i>=0;i--){
		mv1_select_lut_bank(i,bs);
		mv1_load_lut_default(0,bs);
	}
}

static void 
MV1_FreeMasterMem(MV1000 *mv1) {
	int i;
	for(i=0;i<MASTERMEM_PAGES;i++) {
		unsigned long page = mv1->mastermem[i];
		if(page) {
			mem_map_unreserve(virt_to_page(page));
			free_page(page);
		}
	}
}

static int
MV1_AllocMasterMem(MV1000 *mv1) {
	int i;
	for(i=0;i<MASTERMEM_PAGES;i++) {
		unsigned long page = mv1->mastermem[i] = get_free_page(GFP_KERNEL);
		if(page==0) {
			printk("Not enough memory for master DMA\n");
			return -ENOMEM;
		}
		memset((void*)page,0,PAGE_SIZE);
		mem_map_reserve(virt_to_page(page));
	}
	return 0;
}

static int 
mv1_sprint_status(MV1000*mv1,char *buf) 
{
	char *str = buf; 
	str+=sprintf(str,"MuTech MV1000\n");
	return str-buf;
}

static int
mv1000_proc_read (char *page, char **start, off_t off,
                                int count, int *eof,void *data) {
        char * str = page;
	MV1000 *mv1 = (MV1000*)data;
        int len;
        str += mv1_sprint_status(mv1,str);
        len = str - page;
        len -= off;
        if(len < count) {
                *eof = 1;
                if(len<=0)
                        return 0;
        } else
                len = count;
        *start = page+off;
        return len;
}      

static void 
unregister_procinfo(void) {
        int i;
        static char name[20];
        for(i=0;i<MAXBOARDS;i++) {
                sprintf(name,"mv1000/mv1000_%d",i);
                remove_proc_entry(name,0);
        }
        remove_proc_entry("mv1000",0);
}     

static void 
register_procinfo() {
        int i;
        char name[10];
        struct proc_dir_entry *dir;
        struct proc_dir_entry *ent;
        dir = create_proc_entry("mv1000",S_IFDIR|S_IRUGO|S_IXUGO,0);
        for(i=0;i<MAXBOARDS;i++) {
                if(!mv1_board[i])
                        continue;
                sprintf(name,"mv1000_%d",i);
                ent = create_proc_entry(name,S_IFREG|S_IRUGO,dir);
		ent->read_proc = mv1000_proc_read;
                ent->data = (void *)mv1_board[i];
        }
}              
void
MV1_Cleanup(MV1000 *mv1) {
	int i;
	if(mv1==NULL)
		return;
	if(mv1->watchdog_timer_busy) {
		del_timer(&mv1->watchdog_timer);
	}
	unregister_procinfo();
	mv1_set_master_enable(0,mv1);
	mv1_destroy_sdcs(mv1);
	if(mv1->irq) {
		free_irq(mv1->irq,mv1);
	}
	if(mv1->iomem){
		iounmap((void*) mv1->iomem);
	}
	MV1_FreeMasterMem(mv1);
	for(i=0;i<NUM_TLISTS;i++) {
		if(mv1->tlist[i])  
			vfree(mv1->tlist[i]);
	}
	kfree(mv1);
}


void
mv1_cleanup_module(void)
{
	unsigned long flags;
	int i;
        printk("mv1000drv:  cleanup_module called\n");
	save_flags(flags);
	cli();
	if(emutimer_busy) {
		del_timer(&emutimer);
	}
	restore_flags(flags);
        if (MOD_IN_USE)
                printk("mv1000drv: device busy, remove delayed\n");

        if (unregister_chrdev(MV1_MAJOR, "mv1000") != 0)
                printk("cleanup_module failed\n");
        else
                printk("cleanup_module succeeded\n");

        for(i=0;i<MAXBOARDS;i++){
		MV1000 *mv1 = mv1_board[i];
		MV1_Cleanup(mv1);
        }
	if(lpt_port_base) {
		release_region(lpt_port_base,0x3);
	}
}

MV1000 *
MV1_New(u32 pciaddr,u8 irq)  {
	int i;
	MV1000 *mv1 = kmalloc(sizeof(*mv1),GFP_KERNEL);
	if(!mv1)
		return NULL;

	memset(mv1,0,sizeof(*mv1));
	if(MV1_AllocMasterMem(mv1)!=0) {
		MV1_Cleanup(mv1);
		return NULL;	
	}
	for(i=0;i<NUM_TLISTS;i++) {
		mv1->tlist[i] = vmalloc(sizeof(MV1_TransferList));
		if(mv1->tlist[i] == NULL) {
			MV1_Cleanup(mv1);
			return NULL;
		}
	}
	mv1->phys_mem=pciaddr;
  	mv1->iomem = (u_long)ioremap_nocache(mv1->phys_mem,8*1024*1024);
	mv1->vvmem = (u_long) mv1->iomem + 4*1024*1024;
	init_waitqueue_head(&mv1->event_wait);
	init_waitqueue_head(&mv1->frame_end_wait);
	init_waitqueue_head(&mv1->transfer_end_wait);

	for(i=0;i<2;i++) {
        	mv1_init_xilinx(mv1);
		mv1_init_lm1882(mv1);
       		mv1_init_pll(mv1);
       		mv1_init_dacs(mv1);
       		mv1_init_filter(mv1);
       		mv1_init_luts(mv1);
	}
        mv1->i_time = 0;

        mv1->i_time = 1; 
	mv1->snap = 0;
	mv1->snap_flag = 0;
        mv1->curr_gw.sx = 0;
        mv1->curr_gw.sy = 0;
        mv1->curr_gw.dx = 768;
        mv1->curr_gw.dy = 576;
        mv1->curr_gw.startaddr = mv1->grab_start_addr;
        mv1->curr_gw.interlaced = 1;
	mv1->curr_gw.start_even = 1;

	mv1_recalc_curr_gw(mv1);
	mv1_recalc_curr_gw(mv1);

	if(irq && !emulate_irq) {
               	if(request_irq(irq,mv1_interrupt,SA_INTERRUPT,"mv-1000",mv1)){
                       	printk("can't get irq %d\n",irq);
               	} else {
			mv1->irq=irq;
		}
	} else if(emulate_irq) {
		init_timer(&emutimer);
               	emutimer.expires=jiffies+HZ; 
		add_timer(&emutimer);
		emutimer_busy=1;
	}
        cli();
        mv1_set_frame_end_irq(1,mv1);
	mv1_set_transfer_end_irq(1,mv1);
	mv1_clear_all_irqs(mv1);
	sti();

	init_timer(&mv1->watchdog_timer);
	mv1->watchdog_timer.expires=jiffies+HZ;
	mv1->watchdog_timer.function = mv1_watchdog;
	mv1->watchdog_timer_busy = 1;
        mv1->watchdog_timer.data = (unsigned long)mv1;
	add_timer(&mv1->watchdog_timer);

//	printk("New MV1000 at 0x%08x, irq %d, mapped to 0x%08x, 0x%08x\n",pciaddr,irq,mv1->iomem,mv1->vvmem);
	return mv1;
}

/*
 * ------------------------------------------------------------------
 *
 * mv1_pci_find --
 * 	 
 *	find the mv-1000 on the bus. the mv-1000 has only a
 *	configuration space of 64 bytes. the rest isn't decoded.
 *
 * ------------------------------------------------------------------
 */



static void
mv1_pci_find() 
{
	int numboards=0;
	struct pci_dev *dev=NULL;
	if(!pci_present()) {
		return;
	}
	while((dev=pci_find_device(PCI_VENDOR_ID_MUTECH, PCI_DEVICE_ID_MUTECH_MV1000,dev))) {
		MV1000 *mv1;
		u32 iobase;
		u16 cmd;
		u8 irq = dev->irq;
		u8 revision;

		iobase = pci_resource_start(dev,0);
                pci_write_config_byte(dev, PCI_LATENCY_TIMER, 136 );

		pci_read_config_word(dev, PCI_COMMAND, &cmd);
                if(!(cmd & PCI_COMMAND_MASTER)){
                        printk("mv1000: master not enabled, trying to enable it\n");
		} 
                pci_write_config_word(dev, PCI_COMMAND, cmd | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER );
                pci_read_config_byte(dev, PCI_REVISION_ID, &revision);
		mv1_board[numboards] = mv1 = MV1_New(iobase,irq);
		if(mv1!=NULL) {
			numboards++;
			mv1_numboards++;
		} 
        }
}


int
mv1_init_module( void)
{	
	printk( "mv1000drv:  init_module called\n");

	if (register_chrdev(MV1_MAJOR, "mv1000", &mv1_fops)) {
		printk("register_chrdev failed: goodbye world :-(\n");
		return -EIO;
	} 
	handle_triton();
	mv1_pci_find();
	printk("mv1000: %d Board(s) found\n",mv1_numboards);
	if(mv1_numboards==0) {
		mv1_cleanup_module();
		return -ENODEV;
	}
	if(lpt_port_base) {
		if(check_region(lpt_port_base,0x3) < 0 ) {
                	printk("mv-1000: lpt-port 0x%x not available !\n",lpt_port_base);
			lpt_port_base = 0;
        	}
		else {
			request_region(lpt_port_base,0x3,"mv1000(camctl)");
		}	
	}
	register_procinfo();
	init_module_done=1;
	return 0;
}

module_init(mv1_init_module);
module_exit(mv1_cleanup_module); 
